# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL. If not, see <http://www.gnu.org/licenses/>.

"""Factory functions for creating proximal operators.

Functions with ``cconj`` mean the proximal of the convex conjugate and are
provided for convenience.

For more details see :ref:`proximal_operators` and references therein. For
more details on proximal operators including how to evaluate the proximal
operator of a variety of functions see [PB2014]_. """

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import super

import numpy as np
import scipy.special

from odl.operator import (Operator, IdentityOperator, ScalingOperator,
                          ConstantOperator, DiagonalOperator)
from odl.space import ProductSpace
from odl.set import LinearSpaceElement


__all__ = ('combine_proximals', 'proximal_cconj', 'proximal_translation',
           'proximal_arg_scaling', 'proximal_quadratic_perturbation',
           'proximal_composition', 'proximal_const_func',
           'proximal_box_constraint', 'proximal_nonnegativity',
           'proximal_l1', 'proximal_cconj_l1',
           'proximal_l2', 'proximal_cconj_l2',
           'proximal_l2_squared', 'proximal_cconj_l2_squared',
           'proximal_cconj_kl', 'proximal_cconj_kl_cross_entropy')


def combine_proximals(*factory_list):
    """Combine proximal operators into a diagonal product space operator.

    This assumes the functional to be separable across variables in order to
    make use of the separable sum property of proximal operators.

    Parameters
    ----------
    factory_list : sequence of callables
        Proximal operator factories to be combined.

    Returns
    -------
    diag_op : function
        Returns a diagonal product space operator factory to be initialized
        with the same step size parameter

    Notes
    -----
    That two functionals :math:`F` and :math:`G` are separable across variables
    means that :math:`F((x, y)) = F(x)` and :math:`G((x, y)) = G(y)`, and in
    this case the proximal operator of the sum is given by

    .. math::
        \mathrm{prox}_{\\sigma (F(x) + G(y))}(x, y) =
        (\mathrm{prox}_{\\sigma F}(x), \mathrm{prox}_{\\sigma G}(y)).
    """

    def diag_op_factory(sigma):
        """Diagonal matrix of operators.

        Parameters
        ----------
        sigma : positive float
            Step size parameter

        Returns
        -------
        diag_op : `DiagonalOperator`
        """
        return DiagonalOperator(
            *[factory(sigma) for factory in factory_list])

    return diag_op_factory


def proximal_cconj(prox_factory):
    """Calculate the proximal of the dual using Moreau decomposition.

    Parameters
    ----------
    prox_factory : callable
        A factory function that, when called with a step size, returns the
        proximal operator of ``F``

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    The Moreau identity states that for any convex function :math:`F` with
    convex conjugate :math:`F^*`, the proximals satisfy

    .. math::

        \mathrm{prox}_{\\sigma F^*}(x) +\\sigma \,
        \mathrm{prox}_{F / \\sigma}(x / \\sigma) = x

    where :math:`\\sigma` is a scalar step size. Using this, the proximal of
    the convex conjugate is given by

    .. math::

        \mathrm{prox}_{\\sigma F^*}(x) =
        x - \\sigma \, \mathrm{prox}_{F / \\sigma}(x / \\sigma)

    Note that since :math:`(F^*)^* = F`, this can be used to get the proximal
    of the original function from the proximal of the convex conjugate.

    For reference on the Moreau identity, see [CP2011c]_.
    """

    def cconj_prox_factory(sigma):
        """Create proximal for the dual with a given sigma.

        Parameters
        ----------
        sigma : positive float
            Step size parameter

        Returns
        -------
        proximal : `Operator`
            The proximal operator of ``s * F^*`` where ``s`` is the step size
        """
        sigma = float(sigma)
        prox_other = (sigma * prox_factory(1.0 / sigma) * (1.0 / sigma))
        return IdentityOperator(prox_other.domain) - prox_other

    return cconj_prox_factory


def proximal_translation(prox_factory, y):
    """Calculate the proximal of the translated function F(x - y).

    Parameters
    ----------
    prox_factory : callable
        A factory function that, when called with a step size, returns the
        proximal operator of ``F``.
    y : Element in domain of ``F``.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    Given a functional :math:`F`, this is calculated according to the rule

    .. math::

        \mathrm{prox}_{\\sigma  F( \cdot - y)}(x) =
        y + \mathrm{prox}_{\\sigma F}(x - y)

    where :math:`y` is the translation, and :math:`\\sigma` is the step size.

    For reference on the identity used, see [CP2011c]_.
    """

    def translation_prox_factory(sigma):
        """Create proximal for the translation with a given sigma.

        Parameters
        ----------
        sigma : positive float
            Step size parameter

        Returns
        -------
        proximal : `Operator`
            The proximal operator of ``s * F( . - y)`` where ``s`` is the
            step size
        """

        return (ConstantOperator(y) + prox_factory(sigma) *
                (IdentityOperator(y.space) - ConstantOperator(y)))

    return translation_prox_factory


def proximal_arg_scaling(prox_factory, scaling):
    """Calculate the proximal of function F(x * scaling).

    Parameters
    ----------
    prox_factory : callable
        A factory function that, when called with a step size, returns the
        proximal operator of ``F``
    scaling : float
        Scaling parameter

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    Given a functional :math:`F`, this is calculated according to the rule

    .. math::

        \mathrm{prox}_{\\sigma F( \cdot scal)}(x) =
        \\frac{1}{scal} \mathrm{prox}_{\\sigma \, scal^2 \, F }(x \, scal)

    where :math:`scal` is the scaling parameter, and :math:`\\sigma` is the
    step size.

    For reference on the identity used, see [CP2011c]_.
    """

    scaling = float(scaling)
    if scaling == 0:
        return proximal_const_func(prox_factory(1.0).domain)

    def arg_scaling_prox_factory(sigma):
        """Create proximal for the translation with a given sigma.

        Parameters
        ----------
        sigma : positive float
            Step size parameter

        Returns
        -------
        proximal : `Operator`
            The proximal operator of ``sigma * F( . * a)`` where ``sigma`` is
            the step size
        """
        prox = prox_factory(sigma * scaling ** 2)
        return (1 / scaling) * prox * scaling

    return arg_scaling_prox_factory


def proximal_quadratic_perturbation(prox_factory, a, u=None):
    """Calculate the proximal of function F(x) + a * \|x\|^2 + <u,x>.

    Parameters
    ----------
    prox_factory : callable
        A factory function that, when called with a step size, returns the
        proximal operator of ``F``
    a : non-negative float
        Scaling of the quadratic term
    u : Element in domain of F, optional
        Defines the linear functional. For ``None``, the zero element
        is taken.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    Given a functional :math:`F`, this is calculated according to the rule

    .. math::

        \mathrm{prox}_{\\sigma \left(F( \cdot ) + a \| \cdot \|^2 +
        <u, \cdot >\\right)}(x) =
        c \; \mathrm{prox}_{\\sigma F( \cdot \, c)}((x - \\sigma u) c)

    where :math:`c` is the constant

    .. math::

        c = \\frac{1}{\\sqrt{2 \\sigma a + 1}},

    :math:`a` is the scaling parameter belonging to the quadratic term,
    :math:`u` is the space element defining the linear functional, and
    :math:`\\sigma` is the step size.

    For reference on the identity used, see [CP2011c]_. Note that this identity
    is not the exact one given in the reference, but was recalculated for
    arbitrary step lengths.
    """

    a = float(a)
    if a < 0:
        raise ValueError('scaling parameter muts be non-negative, got {}'
                         ''.format(a))

    if u is not None and not isinstance(u, LinearSpaceElement):
        raise TypeError('`u` must be `None` or a `LinearSpaceElement` '
                        'instance, got {!r}.'.format(u))

    def quadratic_perturbation_prox_factory(sigma):
        """Create proximal for the quadratic perturbation with a given sigma.

        Parameters
        ----------
        sigma : positive float
            Step size parameter

        Returns
        -------
        proximal : `Operator`
            The proximal operator of ``sigma * (F(x) + a * \|x\|^2 + <u,x>)``,
            where ``sigma`` is the step size
        """
        const = 1.0 / np.sqrt(sigma * 2.0 * a + 1)
        prox = proximal_arg_scaling(prox_factory, const)(sigma)
        if u is not None:
            return (const * prox *
                    (ScalingOperator(u.space, const) - sigma * const * u))
        else:
            return const * prox * const

    return quadratic_perturbation_prox_factory


def proximal_composition(proximal, operator, mu):
    """Proximal operator factory of functional composed with unitary operator.

    For a functional ``F`` and a linear unitary `Operator` ``L`` this is the
    factory for the proximal operator of ``F * L``.

    Parameters
    ----------
    prox_factory : callable
        A factory function that, when called with a step size returns the
        proximal operator of ``F``
    operator : `Operator`
        The operator to compose the functional with
    mu : ``operator.field`` element
        Scalar such that ``(operator.adjoint * operator)(x) = mu * x``

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    Given a linear operator :math:`L` with the property that for a scalar
    :math:`\\mu`

    .. math::

        L^*(L(x)) = \\mu * x

    and a convex function :math:`F`, the following identity holds

    .. math::

        \mathrm{prox}_{\\sigma F \circ L}(x) = x + \\frac{1}{\\mu}
        L^* \left( \mathrm{prox}_{\\mu \\sigma F}(Lx) - Lx \\right)

    This factory function implements this functionality.

    There is no simple formula for more general operators.

    The function cannot verify that the operator is unitary, the user needs
    to verify this.

    For reference on the identity used, see [CP2011c]_.
    """

    def proximal_composition_factory(sigma):
        """Create proximal for the dual with a given sigma

        Parameters
        ----------
        sigma : positive float
            Step size parameter

        Returns
        -------
        proximal : `Operator`
            The proximal operator of ``prox[sigma * F * L](x)``
        """
        Id = IdentityOperator(operator.domain)
        Ir = IdentityOperator(operator.range)
        prox_muf = proximal(mu * sigma)
        return (Id +
                (1.0 / mu) * operator.adjoint * ((prox_muf - Ir) * operator))

    return proximal_composition_factory


def proximal_const_func(space):
    """Proximal operator factory of the constant functional.

    Function to initialize the proximal operator of the constant functional
    defined on ``space``.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of the functional G=constant

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    The constant functional :math:`G` is defind as :math:`G(x) = constant`
    for all values of :math:`x`. The proximal operator of this functional is
    the identity operator

    .. math::

        \mathrm{prox}_{\\sigma G}(x) = x

    Note that it is independent of :math:`\\sigma`.
    """

    def identity_factory(sigma):
        """Return an instance of the proximal operator.

        Parameters
        ----------
        sigma : positive float
            Unused step size parameter. Introduced to provide a unified
            interface.

        Returns
        -------
        id : `IdentityOperator`
            The proximal operator instance of G = 0 which is the
            identity operator
        """

        return IdentityOperator(space)

    return identity_factory


def proximal_box_constraint(space, lower=None, upper=None):
    """Proximal operator factory for ``G(x) = ind(a <= x <= b)``.

    If P is the set of elements with a <= x <= b, the indicator function of
    which is defined as::

        ind(a <= x <= b) = {0 if x in P, infinity if x is not in P}

    with x being an element in ``space``.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of the functional G(x)
    lower : ``space.field`` element or ``space`` `element-like`, optional
        The lower bound.
        Default: ``None``, interpreted as -infinity
    upper : ``space.field`` element or ``space`` `element-like`, optional
        The upper bound.
        Default: ``None``, interpreted as +infinity

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    If :math:`P` is an interval :math:`[a,b]`, the indicator function is
    defined as

    .. math::

        I_{P}(x) = \\begin{cases}
        0 & \\text{if } x \\in P, \\\\
        \\infty & \\text{if } x \\not \\in P
        \\end{cases}

    For a step size :math:`\\sigma`, the proximal operator of
    :math:`\\sigma I_{P}` is given by the projection onto the interval

    .. math::

         \mathrm{prox}_{\\sigma I_{P}}(x) = \\begin{cases}
         a & \\text{if } x < a, \\\\
         x & \\text{if } x \\in [a,b], \\\\
         b & \\text{if } x > b.
         \\end{cases}

    The proximal operator is independent of :math:`\\sigma` and invariant under
    a positive rescaling of :math:`I_{P}(x)`, since that leaves the indicator
    function unchanged.

    For spaces of the form :math:`R^n`, the definition extends naturally
    in each component.

    See Also
    --------
    proximal_nonnegativity : Special case with ``lower=0, upper=infty``
    """

    # Convert element-likes if needed, also does some space checking
    if lower is not None and lower not in space and lower not in space.field:
        lower = space.element(lower)
    if upper is not None and upper not in space and upper not in space.field:
        upper = space.element(upper)

    if lower in space.field and upper in space.field:
        if lower > upper:
            raise ValueError('invalid values, `lower` ({}) > `upper` ({})'
                             ''.format(lower, upper))

    class ProxOpBoxConstraint(Operator):

        """Proximal operator for G(x) = ind(a <= x <= b)."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
                Step size parameter, not used.
            """
            super().__init__(domain=space, range=space, linear=False)

        def _call(self, x, out):
            """Apply the operator to ``x`` and store the result in ``out``."""

            if lower is not None and upper is None:
                x.ufunc.maximum(lower, out=out)
            elif lower is None and upper is not None:
                x.ufunc.minimum(upper, out=out)
            elif lower is not None and upper is not None:
                x.ufunc.maximum(lower, out=out)
                out.ufunc.minimum(upper, out=out)
            else:
                out.assign(x)

    return ProxOpBoxConstraint


def proximal_nonnegativity(space):
    """Function to create the proximal operator of ``G(x) = ind(x >= 0)``.

    Function for the proximal operator of the functional ``G(x)=ind(x >= 0)``
    to be initialized.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of the functional G(x)

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    See Also
    --------
    proximal_box_constraint
    """

    return proximal_box_constraint(space, lower=0)


def proximal_cconj_l2(space, lam=1, g=None):
    """Proximal operator factory of the convex conj of the l2-norm/distance.

    Function for the proximal operator of the convex conjugate of the
    functional F where F is the l2-norm (or distance to g, if given)::

        F(x) =  lam ||x - g||_2

    with x and g elements in ``space``, scaling factor lam, and given data g.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of F(x). Needs to be a Hilbert space.
        That is, have an inner product (`LinearSpace.inner`).
    g : ``space`` element
        An element in ``space``
    lam : positive float
        Scaling factor or regularization parameter

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    Most problems are forumlated for the squared norm/distance, in that case
    use the `proximal_cconj_l2_squared` instead.

    The :math:`L_2`-norm/distance :math:`F` is given by is given by

    .. math::

        F(x) = \\lambda \|x - g\|_2

    The convex conjugate :math:`F^*` of :math:`F` is given by

    .. math::

        F^*(y) = \\begin{cases}
        0 & \\text{if } \|y-g\|_2 \leq \\lambda, \\\\
        \\infty & \\text{else.}
        \\end{cases}

    For a step size :math:`\\sigma`, the proximal operator of
    :math:`\\sigma F^*` is given by the projection onto the set of :math:`y`
    satisfying :math:`\|y-g\|_2 \leq \\lambda`, i.e., by

    .. math::

        \mathrm{prox}_{\\sigma F^*}(y) = \\begin{cases}
        \\lambda \\frac{y - g}{\|y - g\|}
        & \\text{if } \|y-g\|_2 > \\lambda, \\\\
        y & \\text{if } \|y-g\|_2 \leq \\lambda
        \\end{cases}

    Note that the expression is independent of :math:`\sigma`.

    See Also
    --------
    proximal_l2 : proximal without convex conjugate
    proximal_cconj_l2_squared : proximal for squared norm/distance
    """
    prox_l2 = proximal_l2(space, lam=lam, g=g)
    return proximal_cconj(prox_l2)


def proximal_l2(space, lam=1, g=None):
    """Proximal operator factory of the l2-norm/distance.

    Function for the proximal operator of the functional ``F`` where ``F``
    is the l2-norm (or distance to g, if given)::

        ``F(x) =  lam ||x - g||_2``

    Parameters
    ----------
    space : `LinearSpace`
        Domain of F(x). Needs to be a Hilbert space.
        That is, have an inner product (`LinearSpace.inner`).
    g : ``space`` element
        An element in ``space``.
    lam : positive float
        Scaling factor or regularization parameter.

    Returns
    -------
    prox_factory : callable
        Factory for the proximal operator to be initialized.

    Notes
    -----
    Most problems are forumlated for the squared norm/distance, in that case
    use `proximal_l2_squared` instead.

    The :math:`L_2`-norm/distance :math:`F` is given by

    .. math::

        F(x) = \\lambda \|x - g\|_2

    For a step size :math:`\\sigma`, the proximal operator of :math:`\\sigma F`
    is given by

    .. math::

        \mathrm{prox}_{\\sigma F}(y) = \\begin{cases}
        \\frac{1 - c}{\|y-g\|} \\cdot y  + c \cdot g
        & \\text{if } c < g, \\\\
        g & \\text{else},
        \\end{cases}

    where :math:`c = \\sigma \\frac{\\lambda}{\|y - g\|_2}`.

    See Also
    --------
    proximal_l2_squared : proximal for squared norm/distance
    proximal_cconj_l2 : proximal for convex conjugate
    """
    lam = float(lam)

    if g is not None and g not in space:
        raise TypeError('{!r} is not an element of {!r}'.format(g, space))

    class ProximalL2(Operator):

        """Proximal operator of the l2-norm/distance."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
                Step size parameter
            """
            self.sigma = float(sigma)
            super().__init__(domain=space, range=space, linear=False)

        def _call(self, x, out):
            """Apply the operator to ``x`` and stores the result in ``out``."""

            dtype = getattr(self.domain, 'dtype', float)
            eps = np.finfo(dtype).resolution * 10

            if g is None:
                x_norm = x.norm() * (1 + eps)
                if x_norm > 0:
                    step = self.sigma * lam / x_norm
                else:
                    step = np.infty

                if step < 1.0:
                    out.lincomb(1.0 - step, x)
                else:
                    out.set_zero()

            else:
                x_norm = (x - g).norm() * (1 + eps)
                if x_norm > 0:
                    step = self.sigma * lam / x_norm
                else:
                    step = np.infty

                if step < 1.0:
                    out.lincomb(1.0 - step, x, step, g)
                else:
                    out.assign(g)

    return ProximalL2


def proximal_cconj_l2_squared(space, lam=1, g=None):
    """Proximal operator factory of the convex conj of the squared l2-norm/dist

    Function for the proximal operator of the convex conjugate of the
    functional F where F is the l2-norm (or distance to g, if given)::

        F(x) =  lam ||x - g||_2^2

    with x and g elements in ``space``, scaling factor lam, and given data g.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of F(x). Needs to be a Hilbert space.
        That is, have an inner product (`LinearSpace.inner`).
    g : ``space`` element
        An element in ``space``
    lam : positive float
        Scaling factor or regularization parameter

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    The squared :math:`L_2`-norm/distance :math:`F` is given by

    .. math::

        F(x) =  \\lambda \|x - g\|_2^2.

    The convex conjugate :math:`F^*` of :math:`F` is given by

    .. math::

        F^*(y) = \\frac{1}{4\\lambda} \left( \|
        y\|_2^2 + \langle y, g \\rangle \\right)

    For a step size :math:`\\sigma`, the proximal operator of
    :math:`\\sigma F^*` is given by

    .. math::

        \mathrm{prox}_{\\sigma F^*}(y) = \\frac{y - \\sigma g}{1 +
        \\sigma/(2 \\lambda)}

    See Also
    --------
    proximal_cconj_l2 : proximal without square
    proximal_l2_squared : proximal without convex conjugate
    """
    lam = float(lam)

    if g is not None and g not in space:
        raise TypeError('{!r} is not an element of {!r}'.format(g, space))

    class ProximalCConjL2Squared(Operator):

        """Proximal operator of the convex conj of the squared l2-norm/dist."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
                Step size parameter
            """
            self.sigma = float(sigma)
            super().__init__(domain=space, range=space, linear=g is None)

        def _call(self, x, out):
            """Apply the operator to ``x`` and stores the result in ``out``"""

            # (x - sig*g) / (1 + sig/(2 lam))

            sig = self.sigma
            if g is None:
                out.lincomb(1.0 / (1 + 0.5 * sig / lam), x)
            else:
                out.lincomb(1.0 / (1 + 0.5 * sig / lam), x,
                            -sig / (1 + 0.5 * sig / lam), g)

    return ProximalCConjL2Squared


def proximal_l2_squared(space, lam=1, g=None):
    """Proximal operator factory of the squared l2-norm/distance.

    Function for the proximal operator of the convex conjugate of the
    functional F where F is the l2-norm (or distance to g, if given)::

        F(x) =  lam ||x - g||_2^2

    with x and g elements in ``space``, scaling factor lam, and given data g.

    Parameters
    ----------
    space : `LinearSpace`
        Domain of F(x). Needs to be a Hilbert space.
        That is, have an inner product (`LinearSpace.inner`).
    g : ``space`` element
        An element in ``space``
    lam : positive float
        Scaling factor or regularization parameter

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    The squared :math:`L_2`-norm/distance :math:`F` is given by

    .. math::

        F(x) =  \\lambda \|x - g\|_2^2.

    For a step size :math:`\\sigma`, the proximal operator of :math:`\\sigma F`
    is given by

    .. math::

        \mathrm{prox}_{\\sigma F}(x) = \\frac{x + 2 \\sigma \\lambda g}
        {1 + 2 \\sigma \\lambda}.

    See Also
    --------
    proximal_l2 : proximal without square
    proximal_cconj_l2_squared : proximal for convex conjugate
    """
    # TODO: optimize
    prox_cc_l2_squared = proximal_cconj_l2_squared(space, lam=lam, g=g)
    return proximal_cconj(prox_cc_l2_squared)


def proximal_cconj_l1(space, lam=1, g=None, isotropic=False):
    """Proximal operator factory of the convex conj of the l1-norm/distance.

    Function for the proximal operator of the convex conjugate of the
    functional ``F`` where ``F`` is an l1-norm (or distance to g, if given)::

        F(x) = lam ||x - g||_1

    with x and g elements in ``space`` and scaling factor lam.

    Parameters
    ----------
    space : `LinearSpace` or `ProductSpace` of `LinearSpace` spaces
        Domain of the functional F
    g : ``space`` element
        An element in ``space``
    lam : positive float
        Scaling factor or regularization parameter
    isotropic : bool
        If ``True``, take the vectorial 2-norm point-wise. Otherwise,
        use the vectorial 1-norm. Only available if ``space`` is a
        `ProductSpace`.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    The :math:`L_1`-norm/distance :math:`F` is the functional

    .. math::

        F(x) = \\lambda \|x - g\|_1.

    The convex conjugate :math:`F^*` of :math:`F` is given by the indicator
    function of the set :math:`Q_\\lambda`

    .. math::

        F^*(y) = I_{Q_\\lambda} \left(
        \left| \\frac{y}{\\lambda} \\right| +
        \left\\langle \\frac{y}{\\lambda}, g \\right\\rangle
        \\right)

    where :math:`Q_\\lambda` is a hypercube centered at the origin with
    width :math:`2 \\lambda`.

    For a step size :math:`\\sigma`, the proximal operator of
    :math:`\\sigma F^*` is given by

    .. math::

        \mathrm{prox}_{\\sigma F^*}(y) = \\frac{\\lambda (y - \\sigma g)}{
        \\max(\\lambda, |y - \\sigma g|)}

    An alternative formulation is available for `ProductSpace`'s, in which
    case the ``isotropic`` parameter can be used, giving

    .. math::

        F(x) = \\lambda \| \|x - g\|_2 \|_1

    In this case, the dual is

    .. math::

        F^*(y) = \\lambda I_{Q_\\lambda} \left(
        \left|\left| \\frac{y}{\\lambda} \\right|\\right|_2 +
        \left\langle \\frac{y}{\\lambda}, g \\right\\rangle
        \\right)

    For a step size :math:`\\sigma`, the proximal operator of
    :math:`\\sigma F^*` is given by

    .. math::

        \mathrm{prox}_{\\sigma F^*}(y) = \\frac{\\lambda (y - \\sigma g)}{
        \\max(\\lambda, \|y - \\sigma g\|_2)}

    where :math:`\\max( \cdot , \cdot)` thresholds the lower bound of
    :math:`\|y\|_2` point-wise.

    See Also
    --------
    proximal_l1 : proximal without convex conjugate conjugate
    """
    lam = float(lam)

    if g is not None and g not in space:
        raise TypeError('{!r} is not an element of {!r}'.format(g, space))

    if isotropic and not isinstance(space, ProductSpace):
        raise TypeError('`isotropic` given without productspace `space`({})'
                        ''.format(space))
    if (isotropic and isinstance(space, ProductSpace) and
            not space.is_power_space):
        raise TypeError('`isotropic` given with non-powerspace `space`({})'
                        ''.format(space))

    class ProximalCConjL1(Operator):

        """Proximal operator of the convex conj of the l1-norm/distance."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
                Step size parameter
            """
            # sigma is not used
            self.sigma = float(sigma)
            super().__init__(domain=space, range=space, linear=False)

        def _call(self, x, out):
            """Apply the operator to ``x`` and stores the result in ``out``."""

            # lam * (x - sigma * g) / max(lam, |x - sigma * g|)

            if g is not None:
                diff = x - self.sigma * g
            else:
                diff = x

            if isotropic:
                # Calculate |x| = pointwise 2-norm of x

                tmp = diff[0] ** 2
                sq_tmp = x[0].space.element()
                for x_i in diff[1:]:
                    x_i.multiply(x_i, out=sq_tmp)
                    tmp += sq_tmp
                tmp.ufunc.sqrt(out=tmp)

                # Pointwise maximum of |x| and lambda
                tmp.ufunc.maximum(lam, out=tmp)

                # Global scaling
                tmp /= lam

                # Pointwise division
                for out_i, x_i in zip(out, diff):
                    x_i.divide(tmp, out=out_i)

            else:
                # Calculate |x| = pointwise 2-norm of x
                diff.ufunc.absolute(out=out)

                # Pointwise maximum of |x| and lambda
                out.ufunc.maximum(lam, out=out)

                # Global scaling
                out /= lam

                # Pointwise division
                diff.divide(out, out=out)

    return ProximalCConjL1


def proximal_l1(space, lam=1, g=None, isotropic=False):
    """Proximal operator factory of the l1-norm/distance.

    Function for the proximal operator of the functional F where F is an
    l1-norm (or distance to g, if given)::

        F(x) = lam ||x - g||_1

    with x and g elements in ``space``, and scaling factor lam.

    Parameters
    ----------
    space : `LinearSpace` or `ProductSpace`
        Domain of the functional.
    g : ``space`` element
        An element in ``space``.
    lam : positive float
        Scaling factor or regularization parameter.
    isotropic : bool
        If ``True``, take the vectorial 2-norm point-wise. Otherwise,
        use the vectorial 1-norm. Only available if ``space`` is a
        `ProductSpace`.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized

    Notes
    -----
    The :math:`L_1`-norm/distance :math:`F` is the functional

    .. math::

        F(x) = \\lambda \|x - g\|_1.

    For a step size :math:`\\sigma`, the proximal operator of :math:`\\sigma F`
    is

    .. math::

        \mathrm{prox}_{\\sigma F}(y) = \\begin{cases}
        y - \\sigma \\lambda
        & \\text{if } y > g + \\sigma \\lambda, \\\\
        0
        & \\text{if } g - \\sigma \\lambda \\leq y \\leq g +
        \\sigma \\lambda \\\\
        y + \\sigma \\lambda
        & \\text{if } y < g - \\sigma \\lambda,
        \\end{cases}

    An alternative formulation is available for `ProductSpace`'s, where the
    the ``isotropic`` parameter can be used, giving

    .. math::

        F(x) = \\lambda \| \|x - g\|_2 \|_1

    The proximal can be calculated using the Moreau equality (also known as
    Moreau decomposition or Moreau identity). See for example [BC2011]_.

    See Also
    --------
    proximal_cconj_l1 : proximal for convex conjugate
    """
    # TODO: optimize
    prox_cc_l1 = proximal_cconj_l1(space, lam=lam, g=g, isotropic=isotropic)
    return proximal_cconj(prox_cc_l1)


def proximal_cconj_kl(space, lam=1, g=None):
    """Proximal operator factory of the convex conjugate of the KL divergence.

    Function returning the proximal operator of the convex conjugate of the
    functional F where F is the entropy-type Kullback-Leibler (KL) divergence::

        F(x) = sum_i (x_i - g_i + g_i ln(g_i) - g_i ln(pos(x_i))) + ind_P(x)

    with ``x`` and ``g`` elements in the linear space ``X``, and ``g``
    non-negative. Here, ``pos`` denotes the nonnegative part, and ``ind_P`` is
    the indicator function for nonnegativity.

    Parameters
    ----------
    space : `FnBase`
        Space X which is the domain of the functional F
    g : ``space`` element, optional
        Data term, positive. If None it is take as the one-element.
    lam : positive float
        Scaling factor.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized.

    See Also
    --------
    proximal_cconj_kl_cross_entropy : proximal for releated functional

    Notes
    -----
    The functional is given by the expression

    .. math::

        F(x) = \\sum_i (x_i - g_i + g_i \\ln(g_i) - g_i \\ln(pos(x_i))) +
        I_{x \\geq 0}(x)

    The indicator function :math:`I_{x \geq 0}(x)` is used to restrict the
    domain of :math:`F` such that :math:`F` is defined over whole space
    :math:`X`. The non-negativity thresholding :math:`pos` is used to define
    :math:`F` in the real numbers.

    Note that the functional is not well-defined without a prior g. Hence, if g
    is omitted this will be interpreted as if g is equal to the one-element.

    The convex conjugate :math:`F^*` of :math:`F` is

    .. math::

        F^*(p) = \\sum_i (-g_i \\ln(pos({1_X}_i - p_i))) +
        I_{1_X - p \geq 0}(p)

    where :math:`p` is the variable dual to :math:`x`, and :math:`1_X` is an
    element of the space :math:`X` with all components set to 1.

    The proximal operator of the convex conjugate of F is

    .. math::

        \mathrm{prox}_{\\sigma (\\lambda F)^*}(x) =
        \\frac{\\lambda 1_X + x - \\sqrt{(x -  \\lambda 1_X)^2 +
        4 \\lambda \\sigma g}}{2}

    where :math:`\\sigma` is the step size-like parameter, and :math:`\\lambda`
    is the weighting in front of the function :math:`F`.

    KL based objectives are common in MLEM optimization problems and are often
    used when data noise governed by a multivariate Poisson probability
    distribution is significant.

    The intermediate image estimates can have negative values even though
    the converged solution will be non-negative. Non-negative intermediate
    image estimates can be enforced by adding an indicator function ind_P
    the primal objective.

    This functional :math:`F`, described above, is related to the
    Kullback-Leibler cross entropy functional. The KL cross entropy is the one
    diescribed in `this Wikipedia article
    <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence>`_, and
    the functional :math:`F` is obtained by switching place of the prior and
    the varialbe in the KL cross entropy functional. See the See Also section.
    """
    lam = float(lam)

    if g is not None and g not in space:
        raise TypeError('{} is not an element of {}'.format(g, space))

    class ProximalCConjKL(Operator):

        """Proximal operator of the convex conjugate of the KL divergence."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
            """
            self.sigma = float(sigma)
            super().__init__(domain=space, range=space, linear=False)

        def _call(self, x, out):
            """Apply the operator to ``x`` and stores the result in ``out``."""

            # 1 / 2 (lam_X + x - sqrt((x - lam_X) ^ 2 + 4; lam sigma g)

            # out = x - lam_X
            out.assign(x)
            out -= lam

            # (out)^2
            out.ufunc.square(out=out)

            # out = out + 4 lam sigma g
            # If g is None, it is taken as the one element
            if g is None:
                out += 4.0 * lam * self.sigma
            else:
                out.lincomb(1, out, 4.0 * lam * self.sigma, g)

            # out = sqrt(out)
            out.ufunc.sqrt(out=out)

            # out = x - out
            out.lincomb(1, x, -1, out)

            # out = lam_X + out
            out.lincomb(lam, space.one(), 1, out)

            # out = 1/2 * out
            out /= 2

    return ProximalCConjKL


def proximal_cconj_kl_cross_entropy(space, lam=1, g=None):
    """Proximal factory of the convex conjugate of cross entropy KL divergence.

    Function returning the proximal facotry of the convex conjugate of the
    functional F, where F is the corss entorpy Kullback-Leibler (KL)
    divergence given by::

        F(x) = sum_i (x_i ln(pos(x_i)) - x_i ln(g_i) + g_i - x_i) + ind_P(x)

    with ``x`` and ``g`` in the linear space ``X``, and ``g`` non-negative.
    Here, ``pos`` denotes the nonnegative part, and ``ind_P`` is the indicator
    function for nonnegativity.

    Parameters
    ----------
    space : `FnBase`
        Space X which is the domain of the functional F
    g : ``space`` element, optional
        Data term, positive. If None it is take as the one-element.
    lam : positive float
        Scaling factor.

    Returns
    -------
    prox_factory : function
        Factory for the proximal operator to be initialized.


    See Also
    --------
    proximal_cconj_kl : proximal for related functional

    Notes
    -----
    The functional is given by the expression

    .. math::

        F(x) = \\sum_i (x_i \\ln(pos(x_i)) - x_i \\ln(g_i) + g_i - x_i) +
        I_{x \\geq 0}(x)

    The indicator function :math:`I_{x \geq 0}(x)` is used to restrict the
    domain of :math:`F` such that :math:`F` is defined over whole space
    :math:`X`. The non-negativity thresholding :math:`pos` is used to define
    :math:`F` in the real numbers.

    Note that the functional is not well-defined without a prior g. Hence, if g
    is omitted this will be interpreted as if g is equal to the one-element.

    The convex conjugate :math:`F^*` of :math:`F` is

    .. math::

        F^*(p) = \\sum_i g_i (exp(p_i) - 1)

    where :math:`p` is the variable dual to :math:`x`.

    The proximal operator of the convex conjugate of :math:`F` is

    .. math::

        \mathrm{prox}_{\\sigma (\\lambda F)^*}(x) = x - \\lambda
        W(\\frac{\\sigma}{\\lambda} g e^{x/\\lambda})

    where :math:`\\sigma` is the step size-like parameter, :math:`\\lambda` is
    the weighting in front of the function :math:`F`, and :math:`W` is the
    Lambert W function (see, for example, the
    `Wikipedia article <https://en.wikipedia.org/wiki/Lambert_W_function>`_).

    For real-valued input x, the Lambert :math:`W` function is defined only for
    :math:`x \\geq -1/e`, and it has two branches for values
    :math:`-1/e \\leq x < 0`. However, for inteneded use-cases, where
    :math:`\\lambda` and :math:`g` are positive, the argument of :math:`W`
    will always be positive.

    `Wikipedia article on Kullback Leibler divergence
    <https://en.wikipedia.org/wiki/Kullback%E2%80%93Leibler_divergence>`_.
    For further information about the functional, see for example `this article
    <http://ieeexplore.ieee.org/document/1056144/?arnumber=1056144>`_.

    The KL cross entropy functional :math:`F`, described above, is related to
    another functional functional also know as KL divergence. This functional
    is often used as data discrepancy term in inverse problems, when data is
    corrupted with Poisson noise. This functional is obtained by changing place
    of the prior and the variable. See the See Also section.
    """
    lam = float(lam)

    if g is not None and g not in space:
        raise TypeError('{} is not an element of {}'.format(g, space))

    class ProximalCConjKLCrossEntropy(Operator):

        """Proximal operator of conjugate of cross entropy KL divergence."""

        def __init__(self, sigma):
            """Initialize a new instance.

            Parameters
            ----------
            sigma : positive float
            """
            self.sigma = float(sigma)
            super().__init__(domain=space, range=space, linear=False)

        def _call(self, x, out):
            """Apply the operator to ``x`` and stores the result in ``out``."""

            if g is None:
                # If g is None, it is taken as the one element
                # Different branches of lambertw is not an issue, see Notes
                out.lincomb(1, x, -lam, scipy.special.lambertw(
                    (self.sigma / lam) * np.exp(x / lam)))
            else:
                # Different branches of lambertw is not an issue, see Notes
                out.lincomb(1, x,
                            -lam, scipy.special.lambertw(
                                (self.sigma / lam) * g * np.exp(x / lam)))

    return ProximalCConjKLCrossEntropy


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Parallel beam geometries in 2 and 3 dimensions."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import super

import numpy as np

from odl.tomo.geometry.detector import Flat1dDetector, Flat2dDetector
from odl.tomo.geometry.geometry import Geometry, AxisOrientedGeometry
from odl.tomo.util.utility import euler_matrix, perpendicular_vector


__all__ = ('ParallelGeometry', 'Parallel2dGeometry', 'Parallel3dEulerGeometry',
           'Parallel3dAxisGeometry')


class ParallelGeometry(Geometry):

    """Abstract parallel beam geometry in 2 or 3 dimensions.

    Parallel geometries are characterized by a virtual source at
    infinity, such that a unit vector from a detector point towards
    the source (`det_to_src`) is independent of the location on the
    detector.
    """

    def __init__(self, ndim, apart, detector, det_init_pos):
        """Initialize a new instance.

        Parameters
        ----------
        ndim : {2, 3}
            Number of dimensions of this geometry, i.e. dimensionality
            of the physical space in which this geometry is embedded
        apart : `RectPartition`
            Partition of the angle set
        detector : `Detector`
            The detector to use in this geometry
        det_init_pos : `array-like`
            Initial position of the detector reference point
        """
        super().__init__(ndim, apart, detector)

        if self.ndim not in (2, 3):
            raise ValueError('number of dimensions is {}, expected 2 or 3'
                             ''.format(ndim))

        self._det_init_pos = np.asarray(det_init_pos, dtype='float64')
        if self._det_init_pos.shape != (self.ndim,):
            raise ValueError('initial detector position has shape {}, '
                             'expected ({},)'
                             ''.format(self._det_init_pos.shape, self.ndim))

    @property
    def angles(self):
        """Discrete angles given in this geometry."""
        return self.motion_grid.coord_vectors[0]

    def det_refpoint(self, angles):
        """Return the position of the detector ref. point at ``angles``.

        The reference point is given by a rotation of the initial
        position by ``angles``.

        Parameters
        ----------
        angles : float
            Parameters describing the detector rotation, must be
            contained in `motion_params`.

        Returns
        -------
        point : `numpy.ndarray`, shape (`ndim`,)
            The reference point for the given parameters
        """
        if angles not in self.motion_params:
            raise ValueError('`angles` {} not in the valid range {}'
                             ''.format(angles, self.motion_params))
        return self.rotation_matrix(angles).dot(self._det_init_pos)

    def det_to_src(self, angles, dpar, normalized=True):
        """Direction from a detector location to the source.

        In parallel geometry, this function is independent of the
        detector parameter.

        Since the (virtual) source is infinitely far away, only the
        normalized version is valid.

        Parameters
        ----------
        angles : `array-like`
            Euler angles given in radians, must be contained
            in this geometry's `motion_params`
        dpar : float
            Detector parameters, must be contained in this
            geometry's `det_params`
        normalized : bool, optional
            If ``True``, return the normalized version of the vector.
            For parallel geometry, this is the only sensible option.

        Returns
        -------
        vec : `numpy.ndarray`, shape (`ndim`,)
            Unit vector pointing from the detector to the source

        Raises
        ------
        NotImplementedError
            if ``normalized=False`` is given, since this case is not
            well defined.
        """
        if angles not in self.motion_params:
            raise ValueError('`angles` {} not in the valid range {}'
                             ''.format(angles, self.motion_params))

        if dpar not in self.det_params:
            raise ValueError('`dpar` {} not in the valid range '
                             '{}'.format(dpar, self.det_params))

        if not normalized:
            raise NotImplementedError('non-normalized detector to source is '
                                      'not available in parallel case')

        return self.rotation_matrix(angles).dot(self.detector.normal)


class Parallel2dGeometry(ParallelGeometry):

    """Parallel beam geometry in 2d.

    The motion parameter is the counter-clockwise rotation angle around
    the origin, and the detector is a line detector perpendicular to the
    ray direction.

    In the standard configuration, the detector reference point starts
    at ``(1, 0)``, and the initial detector axis is ``(0, 1)``.
    """

    def __init__(self, apart, dpart, **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        apart : 1-dim. `RectPartition`
            Partition of the angle interval
        dpart : 1-dim. `RectPartition`
            Partition of the detector parameter interval
        det_init_pos : `array-like`, shape ``(2,)``, optional
            Initial position of the detector reference point. The zero
            vector is only allowed if ``det_init_axis`` is explicitly
            given.
            Default: ``(1, 0)``.
        det_init_axis : `array-like` (shape ``(2,)``), optional
            Initial axis defining the detector orientation.
            By default, a normalized `perpendicular_vector` to
            ``det_init_pos`` is used, which is only valid if
            ``det_init_axis`` is not zero.
        """
        self._det_init_pos = kwargs.pop('det_init_pos', (1.0, 0.0))
        self._det_init_axis = kwargs.pop('det_init_axis', None)

        if self.det_init_axis is None:
            if np.linalg.norm(self.det_init_pos) <= 1e-10:
                raise ValueError('`det_init_pos` {} is close to '
                                 'zero. This is only allowed for explicit '
                                 'det_init_axis'
                                 ''.format(self.det_init_pos))

            det_init_axis = perpendicular_vector(self.det_init_pos)

        detector = Flat1dDetector(part=dpart, axis=det_init_axis)
        super().__init__(ndim=2, apart=apart, detector=detector,
                         det_init_pos=self.det_init_pos)

        if self.motion_partition.ndim != 1:
            raise ValueError('`apart` dimension {}, expected 1'
                             ''.format(self.motion_partition.ndim))

    @property
    def det_init_pos(self):
        """Position of the detector reference point at angle=0."""
        return self._det_init_pos

    @property
    def det_init_axis(self):
        """Direction of the detector extent at angle=0."""
        return self._det_init_axis

    def rotation_matrix(self, angle):
        """Return the rotation matrix for ``angle``.

        For an angle ``phi``, the matrix is given by::

            rot(phi) = [[cos(phi), -sin(phi)],
                        [sin(phi), cos(phi)]]

        Parameters
        ----------
        angle : float
            Rotation angle given in radians, must be contained in
            this geometry's `motion_params`

        Returns
        -------
        rot : `numpy.ndarray`, shape (2, 2)
            The rotation matrix mapping the standard basis vectors in
            the fixed ("lab") coordinate system to the basis vectors of
            the local coordinate system of the detector reference point,
            expressed in the fixed system
        """
        if angle not in self.motion_params:
            raise ValueError('`angle` {} not in the valid range {}'
                             ''.format(angle, self.motion_params))
        return euler_matrix(angle)

    def __repr__(self):
        """Return ``repr(self)``."""

        inner_fstr = '\n    {!r},\n    {!r}'

        if not np.allclose(self.det_init_pos, [1, 0]):
            inner_fstr += ',\n    det_init_pos={det_init_pos!r}'

        if not np.allclose(self.det_init_pos,
                           perpendicular_vector(self.det_init_pos)):
            inner_fstr += ',\n    det_init_axis={det_init_axis!r}'

        inner_str = inner_fstr.format(self.motion_partition,
                                      self.det_partition,
                                      det_init_pos=self._det_init_pos,
                                      det_init_axis=self.det_init_axis)
        return '{}({})'.format(self.__class__.__name__, inner_str)


class Parallel3dEulerGeometry(ParallelGeometry):

    """Parallel beam geometry in 3d.

    The motion parameters are two or three Euler angles, and the detector
    is flat and two-dimensional.

    In the standard configuration, the detector reference point starts
    at ``(1, 0, 0)``, and the initial detector axes are
    ``[(0, 1, 0), (0, 0, 1)]``.
    """

    def __init__(self, apart, dpart, **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        apart : 2- or 3-dim. `RectPartition`
            Partition of the angle parameter set
        dpart : 2-dim. `RectPartition`
            Partition of the detector parameter interval
        det_init_pos : `array-like`, shape ``(3,)``, optional
            Initial position of the detector reference point. The zero
            vector is only allowed if ``det_init_axes`` is explicitly
            given.
            Default: ``(1, 0, 0)``
        det_init_axes : 2-tuple of `array-like`'s (shape ``(3,)``), optional
            Initial axes defining the detector orientation.
            By default, a normalized `perpendicular_vector` to
            ``det_init_pos`` is taken as first axis, and the normalized
            cross product of these two as second.
        """
        det_init_pos = kwargs.pop('det_init_pos', (1.0, 0.0, 0.0))
        det_init_axes = kwargs.pop('det_init_axes', None)

        if det_init_axes is None:
            if np.linalg.norm(det_init_pos) <= 1e-10:
                raise ValueError('`det_init_pos` {} is close to '
                                 'zero. This is only allowed for explicit '
                                 '`det_init_axes`.'.format(det_init_pos))

            det_init_axis_0 = perpendicular_vector(det_init_pos)
            det_init_axis_1 = np.cross(det_init_pos, det_init_axis_0)
            det_init_axes = (det_init_axis_0, det_init_axis_1)

        detector = Flat2dDetector(part=dpart, axes=det_init_axes)
        super().__init__(ndim=3, apart=apart, detector=detector,
                         det_init_pos=det_init_pos)

        if self.motion_partition.ndim not in (2, 3):
            raise ValueError('`apart` has dimension {}, expected '
                             '2 or 3'.format(self.motion_partition.ndim))

    def rotation_matrix(self, angles):
        """Matrix defining the detector rotation at ``angles``.

        Parameters
        ----------
        angles : `array-like`
            Angles in radians defining the rotation, must be contained
            in this geometry's ``motion_params``

        Returns
        -------
        rot : `numpy.ndarray`, shape ``(3, 3)``
            The rotation matrix mapping the standard basis vectors in
            the fixed ("lab") coordinate system to the basis vectors of
            the local coordinate system of the detector reference point,
            expressed in the fixed system.
        """
        if angles not in self.motion_params:
            raise ValueError('`angles` {} not in the valid range {}'
                             ''.format(angles, self.motion_params))
        return euler_matrix(*angles)

    def __repr__(self):
        """Return ``repr(self)``."""
        # TODO: det_init_axes
        inner_fstr = '\n    {!r},\n    {!r}'

        if not np.allclose(self.origin_to_det, [1, 0]):
            inner_fstr += ',\n    det_init_pos={det_init_pos!r}'

        inner_str = inner_fstr.format(self.motion_partition,
                                      self.det_partition,
                                      det_init_pos=self._det_init_pos)
        return '{}({})'.format(self.__class__.__name__, inner_str)


class Parallel3dAxisGeometry(ParallelGeometry, AxisOrientedGeometry):

    """Parallel beam geometry in 3d with single rotation axis.

    The motion parameter is the rotation angle around the specified
    axis, and the detector is a flat 2d detector perpendicular to the
    ray direction.

    In the standard configuration, the rotation axis is ``(0, 0, 1)``,
    the detector reference point starts at ``(1, 0, 0)``, and the
    initial detector axes are ``[(0, 1, 0), (0, 0, 1)]``.
    """

    def __init__(self, apart, dpart, axis=[0, 0, 1], **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        apart : 1-dim. `RectPartition`
            Partition of the angle interval
        dpart : 2-dim. `RectPartition`
            Partition of the detector parameter interval
        axis : `array-like`, shape ``(3,)``, optional
            Fixed rotation axis defined by a 3-element vector
        det_init_pos : `array-like`, shape ``(3,)``, optional
            Initial position of the detector reference point. The zero
            vector is only allowed if ``det_init_axes`` is explicitly
            given.
            By default, a `perpendicular_vector` to ``axis`` is used.
        det_init_axes : 2-tuple of `array-like`'s (shape ``(3,)``), optional
            Initial axes defining the detector orientation.
            By default, the normalized cross product of ``axis`` and
            ``det_init_pos`` is used as first axis and ``axis`` as second.
        """
        AxisOrientedGeometry.__init__(self, axis)

        det_init_pos = kwargs.pop('det_init_pos', perpendicular_vector(axis))
        det_init_axes = kwargs.pop('det_init_axes', None)

        if det_init_axes is None:
            if np.linalg.norm(det_init_pos) <= 1e-10:
                raise ValueError('initial detector position {} is close to '
                                 'zero. This is only allowed for explicit '
                                 '`det_init_axes`.'.format(det_init_pos))

            det_init_axis_0 = np.cross(self.axis, det_init_pos)
            det_init_axis_0 /= np.linalg.norm(det_init_axis_0)
            det_init_axes = (det_init_axis_0, axis)

        detector = Flat2dDetector(part=dpart, axes=det_init_axes)
        super().__init__(ndim=3, apart=apart, detector=detector,
                         det_init_pos=det_init_pos)

        if self.motion_partition.ndim != 1:
            raise ValueError('`apart` has dimension {}, expected 1'
                             ''.format(self.motion_partition.ndim))

    def __repr__(self):
        """Return ``repr(self)``."""
        arg_fstr = '\n    {!r},\n    {!r}'
        if not np.allclose(self.axis, [0, 0, 1]):
            arg_fstr += ',\n    axis={axis!r}'

        if not np.allclose(self._det_init_pos,
                           perpendicular_vector(self.axis)):
            arg_fstr += ',\n    det_init_pos={det_init_pos!r}'

        default_axes = [np.cross(self.axis, self._det_init_pos), self.axis]
        if not np.allclose(self.detector.axes, default_axes):
            arg_fstr += ',\n    det_init_axes={det_init_axes!r}'

        arg_str = arg_fstr.format(self.motion_partition,
                                  self.det_partition,
                                  axis=self.axis,
                                  det_init_pos=self._det_init_pos,
                                  det_init_axes=self.detector.axes)
        return '{}({})'.format(self.__class__.__name__, arg_str)

    # Fix for bug in ABC thinking this is abstract
    rotation_matrix = AxisOrientedGeometry.rotation_matrix


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Cartesian products of `LinearSpace` instances."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from builtins import range, str, super, zip
from future import standard_library
standard_library.install_aliases()

from numbers import Integral
from itertools import product
import numpy as np

from odl.set import LinearSpace, LinearSpaceElement
from odl.space.weighting import (
    WeightingBase, VectorWeightingBase, ConstWeightingBase, NoWeightingBase,
    CustomInnerProductBase, CustomNormBase, CustomDistBase)
from odl.util.ufuncs import ProductSpaceUFuncs
from odl.util.utility import is_real_dtype


__all__ = ('ProductSpace', 'ProductSpaceElement')


class ProductSpace(LinearSpace):

    """Cartesian product of `LinearSpace`'s.

    A product space is the Cartesian product ``X_1 x ... x X_n`` of
    linear spaces ``X_i``. It is itself a linear space, where the linear
    combination is defined component-wise. Inner product, norm and
    distance can also be defined in natural ways from the corresponding
    functions in the individual components.
    """

    def __init__(self, *spaces, **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        space1,...,spaceN : `LinearSpace` or int
            The individual spaces ("factors / parts") in the product
            space. Can also be given as ``space, n`` with ``n`` integer,
            in which case the power space ``space ** n`` is created.
        exponent : non-zero float or ``float('inf')``, optional
            Order of the product distance/norm, i.e.

            ``dist(x, y) = np.linalg.norm(x-y, ord=exponent)``

            ``norm(x) = np.linalg.norm(x, ord=exponent)``

            Values ``0 <= exponent < 1`` are currently unsupported
            due to numerical instability. See ``Notes`` for further
            information about the interpretation of the values.

            Default: 2.0

        field : `Field`, optional
            Scalar field of the resulting space.
            Default: ``spaces[0].field``

        weight : optional
            Use weighted inner product, norm, and dist. The following
            types are supported as ``weight``:

            ``None`` : no weighting (default)

            `WeightingBase` : weighting class, used directly. Such a
            class instance can be retrieved from the space by the
            `ProductSpace.weighting` property.

            `array-like` : weigh each component with one entry from the
            array. The array must be one-dimensional and have the same
            length as the number of spaces.

            float : same weighting factor in each component

        Other Parameters
        ----------------
        dist : callable, optional
            The distance function defining a metric on the space.
            It must accept two `ProductSpaceElement` arguments and
            fulfill the following mathematical conditions for any
            three space elements ``x, y, z``:

            - ``dist(x, y) >= 0``
            - ``dist(x, y) = 0``  if and only if  ``x = y``
            - ``dist(x, y) = dist(y, x)``
            - ``dist(x, y) <= dist(x, z) + dist(z, y)``

            By default, ``dist(x, y)`` is calculated as ``norm(x - y)``.
            This creates an intermediate array ``x - y``, which can be
            avoided by choosing ``dist_using_inner=True``.

            Cannot be combined with: ``weight, norm, inner``

        norm : callable, optional
            The norm implementation. It must accept an
            `ProductSpaceElement` argument, return a float and satisfy the
            following conditions for all space elements ``x, y`` and scalars
            ``s``:

            - ``||x|| >= 0``
            - ``||x|| = 0``  if and only if  ``x = 0``
            - ``||s * x|| = |s| * ||x||``
            - ``||x + y|| <= ||x|| + ||y||``

            By default, ``norm(x)`` is calculated as ``inner(x, x)``.

            Cannot be combined with: ``weight, dist, inner``

        inner : callable, optional
            The inner product implementation. It must accept two
            `ProductSpaceElement` arguments, return a element from
            the field of the space (real or complex number) and
            satisfy the following conditions for all space elements
            ``x, y, z`` and scalars ``s``:

            - ``<x, y> = conj(<y, x>)``
            - ``<s*x + y, z> = s * <x, z> + <y, z>``
            - ``<x, x> = 0``  if and only if  ``x = 0``

            Cannot be combined with: ``weight, dist, norm``

        dist_using_inner : bool, optional
            Calculate ``dist`` using the formula

                ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``

            This avoids the creation of new arrays and is thus faster
            for large arrays. On the downside, it will not evaluate to
            exactly zero for equal (but not identical) ``x`` and ``y``.

            This option can only be used if ``exponent`` is 2.0.

            Default: ``False``.

            Cannot be combined with: ``dist``

        Returns
        -------
        prodspace : `ProductSpace`

        See Also
        --------
        ProductSpaceElementWeighting
        ProductSpaceConstWeighting

        Examples
        --------
        >>> r2x3 = ProductSpace(odl.rn(2), odl.rn(3))

        Notes
        -----
        Inner product, norm and distance are evaluated by collecting
        the result of the corresponding operation in the individual
        components and reducing the resulting vector to a single number.
        The ``exponent`` parameter influences only this last part,
        not the computations in the individual components. We give the
        exact definitions in the following:

        Let :math:`\mathcal{X} = \mathcal{X}_1 \\times \dots \\times
        \mathcal{X}_d` be a product space, and
        :math:`\langle \cdot, \cdot\\rangle_i`,
        :math:`\lVert \cdot \\rVert_i`, :math:`d_i(\cdot, \cdot)` be
        inner products, norms and distances in the respective
        component spaces.

        **Inner product:**

            :math:`\langle x, y \\rangle =
            \\sum_{i=1}^d \langle x_i, y_i \\rangle_i`

        **Norm:**

        - :math:`p < \infty`:

            :math:`\lVert x\\rVert =
            \left( \sum_{i=1}^d \lVert x_i \\rVert_i^p \\right)^{1/p}`

        - :math:`p = \infty`:

            :math:`\lVert x\\rVert =
            \max_i \lVert x_i \\rVert_i`

        **Distance:**

        - :math:`p < \infty`:

            :math:`d(x, y) =
            \left( \sum_{i=1}^d d_i(x_i, y_i)^p \\right)^{1/p}`

        - :math:`p = \infty`:

            :math:`d(x, y) =
            \max_i d_i(x_i, y_i)`

        To implement own versions of these functions, you can use
        the following snippet to gather the vector of norms (analogously
        for inner products and distances)::

            norms = np.fromiter(
                (xi.norm() for xi in x),
                dtype=np.float64, count=len(x))

        """
        field = kwargs.pop('field', None)
        dist = kwargs.pop('dist', None)
        norm = kwargs.pop('norm', None)
        inner = kwargs.pop('inner', None)
        weight = kwargs.pop('weight', None)
        exponent = kwargs.pop('exponent', 2.0)
        dist_using_inner = bool(kwargs.pop('dist_using_inner', False))
        if kwargs:
            raise TypeError('got unexpected keyword arguments: {}'
                            ''.format(kwargs))

        # Check validity of option combination (3 or 4 out of 4 must be None)
        if sum(x is None for x in (dist, norm, inner, weight)) < 3:
            raise ValueError('invalid combination of options weight, '
                             'dist, norm and inner')

        if any(x is not None for x in (dist, norm, inner)) and exponent != 2.0:
            raise ValueError('`exponent` cannot be used together with '
                             'inner, norm or dist')

        # Make a power space if the second argument is an integer
        if (len(spaces) == 2 and
                isinstance(spaces[0], LinearSpace) and
                isinstance(spaces[1], Integral)):
            spaces = [spaces[0]] * spaces[1]

        # Validate the space arguments
        wrong_spaces = [spc for spc in spaces
                        if not isinstance(spc, LinearSpace)]
        if wrong_spaces:
            raise TypeError('{!r} not LinearSpace instance(s)'
                            ''.format(wrong_spaces))

        if not all(spc.field == spaces[0].field for spc in spaces):
            raise ValueError('all spaces must have the same field')

        # Assign spaces and field
        self.__spaces = tuple(spaces)
        self.__size = len(spaces)
        if field is None:
            if self.size == 0:
                raise ValueError('no spaces provided, cannot deduce field')
            field = self.spaces[0].field

        # Cache for efficiency
        self.__is_power_space = all(spc == self.spaces[0]
                                    for spc in self.spaces[1:])

        super().__init__(field)

        # Assign weighting
        if weight is not None:
            if isinstance(weight, WeightingBase):
                self.__weighting = weight
            elif np.isscalar(weight):
                self.__weighting = ProductSpaceConstWeighting(
                    weight, exponent, dist_using_inner=dist_using_inner)
            elif weight is None:
                # Need to wait until dist, norm and inner are handled
                pass
            else:  # last possibility: make a product space element
                arr = np.asarray(weight)
                if arr.dtype == object:
                    raise ValueError('invalid weight argument {}'
                                     ''.format(weight))
                if arr.ndim == 1:
                    self.__weighting = ProductSpaceElementWeighting(
                        arr, exponent, dist_using_inner=dist_using_inner)
                else:
                    raise ValueError('weighting array has {} dimensions, '
                                     'expected 1'.format(arr.ndim))

        elif dist is not None:
            self.__weighting = ProductSpaceCustomDist(dist)
        elif norm is not None:
            self.__weighting = ProductSpaceCustomNorm(norm)
        elif inner is not None:
            self.__weighting = ProductSpaceCustomInnerProduct(inner)
        else:  # all None -> no weighing
            self.__weighting = ProductSpaceNoWeighting(
                exponent, dist_using_inner=dist_using_inner)

    @property
    def size(self):
        """Number of factors."""
        return self.__size

    def __len__(self):
        """Return ``len(self)``."""
        return self.size

    @property
    def shape(self):
        """Number of spaces per axis."""
        # Currently supporting only 1d product spaces
        return (self.size,)

    @property
    def spaces(self):
        """A tuple containing all spaces."""
        return self.__spaces

    @property
    def is_power_space(self):
        """``True`` if all member spaces are equal."""
        return self.__is_power_space

    @property
    def exponent(self):
        """Exponent of the product space norm/dist, ``None`` for custom."""
        return self.weighting.exponent

    @property
    def weighting(self):
        """This space's weighting scheme."""
        return self.__weighting

    @property
    def is_weighted(self):
        """Return ``True`` if weighting is not `ProductSpaceNoWeighting`."""
        return not isinstance(self.weighting, ProductSpaceNoWeighting)

    def element(self, inp=None, cast=True):
        """Create an element in the product space.

        Parameters
        ----------
        inp : optional
            If ``inp`` is ``None``, a new element is created from
            scratch by allocation in the spaces. If ``inp`` is
            already an element of this space, it is re-wrapped.
            Otherwise, a new element is created from the
            components by calling the ``element()`` methods
            in the component spaces.
        cast : bool
            If ``True``, casting is allowed. Otherwise, a ``TypeError``
            is raised for input that is not a sequence of elements of
            the spaces that make up this product space.

        Returns
        -------
        element : `ProductSpaceElement`
            The new element

        Examples
        --------
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> vec_2, vec_3 = r2.element(), r3.element()
        >>> r2x3 = ProductSpace(r2, r3)
        >>> vec_2x3 = r2x3.element()
        >>> vec_2.space == vec_2x3[0].space
        True
        >>> vec_3.space == vec_2x3[1].space
        True

        Create an element of the product space

        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> prod = ProductSpace(r2, r3)
        >>> x2 = r2.element([1, 2])
        >>> x3 = r3.element([1, 2, 3])
        >>> x = prod.element([x2, x3])
        >>> print(x)
        {[1.0, 2.0], [1.0, 2.0, 3.0]}
        """

        # If data is given as keyword arg, prefer it over arg list
        if inp is None:
            inp = [space.element() for space in self.spaces]

        if inp in self:
            return inp

        if (all(isinstance(v, LinearSpaceElement) and v.space == space
                for v, space in zip(inp, self.spaces))):
            parts = list(inp)
        elif cast:
            # Delegate constructors
            parts = [space.element(arg)
                     for arg, space in zip(inp, self.spaces)]
        else:
            raise TypeError('input {!r} not a sequence of elements of the '
                            'component spaces'.format(inp))

        return self.element_type(self, parts)

    @property
    def examples(self):
        """Return examples from all sub-spaces."""
        for examples in product(*[spc.examples for spc in self.spaces]):
            name = ', '.join(name for name, _ in examples)
            element = self.element([elem for _, elem in examples])
            yield (name, element)

    def zero(self):
        """Create the zero element of the product space.

        The i-th component of the product space zero element is the
        zero element of the i-th space in the product.

        Parameters
        ----------
        None

        Returns
        -------
        zero : ProductSpaceElement
            The zero element in the product space.

        Examples
        --------
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> zero_2, zero_3 = r2.zero(), r3.zero()
        >>> r2x3 = ProductSpace(r2, r3)
        >>> zero_2x3 = r2x3.zero()
        >>> zero_2 == zero_2x3[0]
        True
        >>> zero_3 == zero_2x3[1]
        True
        """
        return self.element([space.zero() for space in self.spaces])

    def one(self):
        """Create the one element of the product space.

        The i-th component of the product space one element is the
        one element of the i-th space in the product.

        Parameters
        ----------
        None

        Returns
        -------
        one : ProductSpaceElement
            The one element in the product space.

        Examples
        --------
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> one_2, one_3 = r2.one(), r3.one()
        >>> r2x3 = ProductSpace(r2, r3)
        >>> one_2x3 = r2x3.one()
        >>> one_2 == one_2x3[0]
        True
        >>> one_3 == one_2x3[1]
        True
        """
        return self.element([space.one() for space in self.spaces])

    def _lincomb(self, a, x, b, y, out):
        """Linear combination ``out = a*x + b*y``."""
        for space, xp, yp, outp in zip(self.spaces, x.parts, y.parts,
                                       out.parts):
            space._lincomb(a, xp, b, yp, outp)

    def _dist(self, x1, x2):
        """Distance between two elements."""
        return self.weighting.dist(x1, x2)

    def _norm(self, x):
        """Norm of an element."""
        return self.weighting.norm(x)

    def _inner(self, x1, x2):
        """Inner product of two elements."""
        return self.weighting.inner(x1, x2)

    def _multiply(self, x1, x2, out):
        """Product ``out = x1 * x2``."""
        for spc, xp, yp, outp in zip(self.spaces, x1.parts, x2.parts,
                                     out.parts):
            spc._multiply(xp, yp, outp)

    def _divide(self, x1, x2, out):
        """Quotient ``out = x1 / x2``."""
        for spc, xp, yp, outp in zip(self.spaces, x1.parts, x2.parts,
                                     out.parts):
            spc._divide(xp, yp, outp)

    def __eq__(self, other):
        """Return ``self == other``.

        Returns
        -------
        equals : bool
            ``True`` if ``other`` is a `ProductSpace` instance, has
            the same length and the same factors. ``False`` otherwise.

        Examples
        --------
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> rn, rm = odl.rn(2), odl.rn(3)
        >>> r2x3, rnxm = ProductSpace(r2, r3), ProductSpace(rn, rm)
        >>> r2x3 == rnxm
        True
        >>> r3x2 = ProductSpace(r3, r2)
        >>> r2x3 == r3x2
        False
        >>> r5 = ProductSpace(*[odl.rn(1)]*5)
        >>> r2x3 == r5
        False
        >>> r5 = odl.rn(5)
        >>> r2x3 == r5
        False
        """
        if other is self:
            return True
        else:
            return (isinstance(other, ProductSpace) and
                    self.shape == other.shape and
                    self.weighting == other.weighting and
                    all(x == y for x, y in zip(self.spaces,
                                               other.spaces)))

    def __getitem__(self, indices):
        """Return ``self[indices]``."""
        if isinstance(indices, Integral):
            return self.spaces[indices]
        elif isinstance(indices, slice):
            return ProductSpace(*self.spaces[indices],
                                field=self.field)
        else:
            return ProductSpace(*[self.spaces[i] for i in indices],
                                field=self.field)

    def __str__(self):
        """Return ``str(self)``."""
        if self.size == 0:
            return '{}'
        elif all(self.spaces[0] == space for space in self.spaces):
            return '{' + str(self.spaces[0]) + '}^' + str(self.size)
        else:
            return ' x '.join(str(space) for space in self.spaces)

    def __repr__(self):
        """Return ``repr(self)``."""
        if self.size == 0:
            return '{}(field={})'.format(self.__class__.__name__, self.field)
        elif self.is_power_space:
            return '{}({!r}, {})'.format(self.__class__.__name__,
                                         self.spaces[0], self.size)
        else:
            inner_str = ', '.join(repr(space) for space in self.spaces)
            return '{}({})'.format(self.__class__.__name__, inner_str)

    @property
    def element_type(self):
        """`ProductSpaceElement`"""
        return ProductSpaceElement


class ProductSpaceElement(LinearSpaceElement):

    """Elements of a `ProductSpace`."""

    def __init__(self, space, parts):
        """Initialize a new instance."""
        super().__init__(space)
        self._parts = list(parts)

    @property
    def parts(self):
        """Parts of this product space element."""
        return self._parts

    @property
    def size(self):
        """Number of factors of this element's space."""
        return self.space.size

    def __len__(self):
        """Return ``len(self)``."""
        return len(self.space)

    def __eq__(self, other):
        """Return ``self == other``.

        Overrides the default `LinearSpace` method since it is
        implemented with the distance function, which is prone to
        numerical errors. This function checks equality per
        component.
        """
        if other is self:
            return True
        elif other not in self.space:
            return False
        else:
            return all(sp == op for sp, op in zip(self.parts, other.parts))

    def __getitem__(self, indices):
        """Return ``self[indices]``."""
        if isinstance(indices, Integral):
            return self.parts[indices]
        elif isinstance(indices, slice):
            return self.space[indices].element(self.parts[indices])
        else:
            out_parts = [self.parts[i] for i in indices]
            return self.space[indices].element(out_parts)

    def __setitem__(self, indices, values):
        """Implement ``self[indices] = values``."""
        try:
            self.parts[indices] = values
        except TypeError:
            for i, index in enumerate(indices):
                self.parts[index] = values[i]

    @property
    def ufunc(self):
        """`ProductSpaceUFuncs`, access to Numpy style ufuncs.

        These are always available if the underlying spaces are
        `NtuplesBase`.

        Examples
        --------
        >>> r22 = odl.ProductSpace(odl.rn(2), 2)
        >>> x = r22.element([[1, -2], [-3, 4]])
        >>> x.ufunc.absolute()
        ProductSpace(rn(2), 2).element([
            [1.0, 2.0],
            [3.0, 4.0]
        ])

        These functions can also be used with non-vector arguments and
        support broadcasting, per component and even recursively:

        >>> x.ufunc.add([1, 2])
        ProductSpace(rn(2), 2).element([
            [2.0, 0.0],
            [-2.0, 6.0]
        ])
        >>> x.ufunc.subtract(1)
        ProductSpace(rn(2), 2).element([
            [0.0, -3.0],
            [-4.0, 3.0]
        ])

        There is also support for various reductions (sum, prod, min, max):

        >>> x.ufunc.sum()
        0.0

        Writing to ``out`` is also supported:

        >>> y = r22.element()
        >>> result = x.ufunc.absolute(out=y)
        >>> result
        ProductSpace(rn(2), 2).element([
            [1.0, 2.0],
            [3.0, 4.0]
        ])
        >>> result is y
        True

        See Also
        --------
        odl.util.ufuncs.NtuplesBaseUFuncs
            Base class for ufuncs in `NtuplesBase` spaces, subspaces may
            override this for greater efficiency.
        odl.util.ufuncs.ProductSpaceUFuncs
            For a list of available ufuncs.
        """
        return ProductSpaceUFuncs(self)

    def __str__(self):
        """Return ``str(self)``."""
        inner_str = ', '.join(str(part) for part in self.parts)
        return '{{{}}}'.format(inner_str)

    def __repr__(self):
        """Return ``repr(self)``.

        Examples
        --------
        >>> from odl import rn  # need to import rn into namespace
        >>> r2, r3 = odl.rn(2), odl.rn(3)
        >>> r2x3 = ProductSpace(r2, r3)
        >>> x = r2x3.element([[1, 2], [3, 4, 5]])
        >>> eval(repr(x)) == x
        True

        The result is readable:

        >>> x
        ProductSpace(rn(2), rn(3)).element([
            [1.0, 2.0],
            [3.0, 4.0, 5.0]
        ])

        Nestled spaces work as well

        >>> X = ProductSpace(r2x3, r2x3)
        >>> x = X.element([[[1, 2], [3, 4, 5]],[[1, 2], [3, 4, 5]]])
        >>> eval(repr(x)) == x
        True
        >>> x
        ProductSpace(ProductSpace(rn(2), rn(3)), 2).element([
            [
                [1.0, 2.0],
                [3.0, 4.0, 5.0]
            ],
            [
                [1.0, 2.0],
                [3.0, 4.0, 5.0]
            ]
        ])
        """
        inner_str = '[\n'
        if len(self) < 5:
            inner_str += ',\n'.join('{}'.format(
                _indent(_strip_space(part))) for part in self.parts)
        else:
            inner_str += ',\n'.join('{}'.format(
                _indent(_strip_space(part))) for part in self.parts[:3])
            inner_str += ',\n    ...\n'
            inner_str += ',\n'.join('{}'.format(
                _indent(_strip_space(part))) for part in self.parts[-1:])

        inner_str += '\n]'

        return '{!r}.element({})'.format(self.space, inner_str)

    def show(self, title=None, indices=None, **kwargs):
        """Display the parts of this product space element graphically.

        Parameters
        ----------
        title : string
            Title of the figures

        indices : index expression, optional
            Indices can refer to parts of a `ProductSpaceElement` and slices
            in the parts in the following way:

            Single index (``indices=0``)
            => display that part

            Single slice (``indices=slice(None)``), or
            index list (``indices=[0, 1, 3]``)
            => display those parts

            Any tuple, for example:
            Created by `numpy.s_` ``indices=np.s_[0, :, :]`` or
            Using a raw tuple ``indices=([0, 3], slice(None))``
            => take the first elements to select the parts and
            pass the rest on to the underlying show methods.

        kwargs
            Additional arguments passed on to the ``show`` methods of
            the parts.

        Returns
        -------
        fig : list of `matplotlib.figure.Figure`
            The resulting figures. It is also shown to the user.

        See Also
        --------
        odl.discr.lp_discr.DiscreteLpElement.show :
            Display of a discretized function
        odl.space.base_ntuples.NtuplesBaseVector.show :
            Display of sequence type data
        odl.util.graphics.show_discrete_data :
            Underlying implementation
        """
        if title is None:
            title = 'ProductSpaceElement'

        if indices is None:
            if len(self) < 5:
                indices = np.arange(self.size)
            else:
                indices = np.linspace(0, self.size - 1, 4, dtype=int)
        else:
            if isinstance(indices, tuple):
                indices, kwargs['indices'] = indices[0], indices[1:]

            if isinstance(indices, slice):
                indices = range(*indices.indices(self.size))
            elif isinstance(indices, Integral):
                indices = [indices]

            # else try with indices as is

        in_figs = kwargs.pop('fig', None)
        in_figs = [None] * len(indices) if in_figs is None else in_figs

        figs = []
        for i, part, fig in zip(indices, self[indices], in_figs):
            fig = part.show(title='{}. Part {}'.format(title, i), fig=fig,
                            **kwargs)
            figs += [fig]

        return figs


# --- Add arithmetic operators that broadcast ---
def _broadcast_arithmetic(op):
    """Return ``op(self, other)`` with broadcasting.

    Parameters
    ----------
    op : string
        Name of the operator, e.g. ``'__add__'``.

    Returns
    -------
    broadcast_arithmetic_op : function
        Function intended to be used as a method for `ProductSpaceVector`
        which performs broadcasting if possible.

    Notes
    -----
    Broadcasting is the operation of "applying an operator multiple times" in
    some sense. For example:

    .. math::
        (1, 2) + 1 = (2, 3)

    is a form of broadcasting. In this implementation, we only allow "single
    layer" broadcasting, i.e., we do not support broadcasting over several
    product spaces at once.
    """
    def _broadcast_arithmetic_impl(self, other):
        if (self.space.is_power_space and other in self.space[0]):
            results = []
            for xi in self:
                res = getattr(xi, op)(other)
                if res is NotImplemented:
                    return NotImplemented
                else:
                    results.append(res)

            return self.space.element(results)
        else:
            return getattr(LinearSpaceElement, op)(self, other)

    # Set docstring
    docstring = """Broadcasted {op}.""".format(op=op)
    _broadcast_arithmetic_impl.__doc__ = docstring

    return _broadcast_arithmetic_impl


for op in ['add', 'sub', 'mul', 'div', 'truediv']:
    for modifier in ['', 'r', 'i']:
        name = '__{}{}__'.format(modifier, op)
        setattr(ProductSpaceElement, name, _broadcast_arithmetic(name))


class ProductSpaceElementWeighting(VectorWeightingBase):

    """Vector weighting for `ProductSpace`.

    For exponent 2.0, a new weighted inner product with vector ``w``
    is defined as::

        <x, y>_w = <w * x, y>

    with component-wise multiplication ``w * x``. For other exponents,
    only ``norm`` and ```dist`` are defined. In the case of exponent
    ``inf``, the weighted norm is::

        ||x||_{w,inf} = ||w * x||_inf

    otherwise it is::

        ||x||_{w,p} = ||w^(1/p) * x||_p

    Not that this definition does **not** fulfill the limit property
    in ``p``, i.e.::

        ||x||_{w,p} --/-> ||x||_{w,inf}  for p --> inf

    unless ``w = (1,...,1)``.

    The vector may only have positive entries, otherwise it does not
    define an inner product or norm, respectively. This is not checked
    during initialization.
    """

    def __init__(self, vector, exponent=2.0, dist_using_inner=False):
        """Initialize a new instance.

        Parameters
        ----------
        vector : 1-dim. `array-like`
            Weighting vector of the inner product.
        exponent : positive float, optional
            Exponent of the norm. For values other than 2.0, no inner
            product is defined.
        dist_using_inner : bool, optional
            Calculate ``dist`` using the formula

                ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``.

            This avoids the creation of new arrays and is thus faster
            for large arrays. On the downside, it will not evaluate to
            exactly zero for equal (but not identical) ``x`` and ``y``.

            Can only be used if ``exponent`` is 2.0.
        """
        super().__init__(vector, impl='numpy', exponent=exponent,
                         dist_using_inner=dist_using_inner)

    def inner(self, x1, x2):
        """Calculate the vector weighted inner product of two elements.

        Parameters
        ----------
        x1, x2 : `ProductSpaceElement`
            Elements whose inner product is calculated.

        Returns
        -------
        inner : float or complex
            The inner product of the two provided elements.
        """
        if self.exponent != 2.0:
            raise NotImplementedError('no inner product defined for '
                                      'exponent != 2 (got {})'
                                      ''.format(self.exponent))

        inners = np.fromiter(
            (x1i.inner(x2i) for x1i, x2i in zip(x1, x2)),
            dtype=x1[0].space.dtype, count=len(x1))

        inner = np.dot(inners, self.vector)
        if is_real_dtype(x1[0].dtype):
            return float(inner)
        else:
            return complex(inner)

    def norm(self, x):
        """Calculate the vector-weighted norm of an element.

        Parameters
        ----------
        x : `ProductSpaceElement`
            Element whose norm is calculated.

        Returns
        -------
        norm : float
            The norm of the provided element.
        """
        if self.exponent == 2.0:
            norm_squared = self.inner(x, x).real  # TODO: optimize?!
            return np.sqrt(norm_squared)
        else:
            norms = np.fromiter(
                (xi.norm() for xi in x), dtype=np.float64, count=len(x))
            if self.exponent in (1.0, float('inf')):
                norms *= self.vector
            else:
                norms *= self.vector ** (1.0 / self.exponent)

            return float(np.linalg.norm(norms, ord=self.exponent))


class ProductSpaceConstWeighting(ConstWeightingBase):

    """Constant weighting for `ProductSpace`.

    For exponent 2.0, a new weighted inner product with constant
    ``c`` is defined as::

        <x, y>_c = c * <x, y>

    For other exponents, only ``norm`` and ```dist`` are defined.
    In the case of exponent ``inf``, the weighted norm is::

        ||x||_{c,inf} = c * ||x||_inf

    otherwise it is::

        ||x||_{c,p} = c^(1/p) * ||x||_p

    Note that this definition does **not** fulfill the limit property
    in ``p``, i.e.::

        ||x||_{c,p} --/-> ||x||_{c,inf}  for p --> inf

    unless ``c = 1``.

    The constant must be positive, otherwise it does not define an
    inner product or norm, respectively.
    """

    def __init__(self, constant, exponent=2.0, dist_using_inner=False):
        """Initialize a new instance.

        Parameters
        ----------
        constant : positive float
            Weighting constant of the inner product
        exponent : positive float, optional
            Exponent of the norm. For values other than 2.0, no inner
            product is defined.
        dist_using_inner : bool, optional
            Calculate ``dist`` using the formula

                ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``

            This avoids the creation of new arrays and is thus faster
            for large arrays. On the downside, it will not evaluate to
            exactly zero for equal (but not identical) ``x`` and ``y``.

            Can only be used if ``exponent`` is 2.0.
        """
        super().__init__(constant, impl='numpy', exponent=exponent,
                         dist_using_inner=dist_using_inner)

    def inner(self, x1, x2):
        """Calculate the constant-weighted inner product of two elements.

        Parameters
        ----------
        x1, x2 : `ProductSpaceElement`
            Elements whose inner product is calculated.

        Returns
        -------
        inner : float or complex
            The inner product of the two provided elements.
        """
        if self.exponent != 2.0:
            raise NotImplementedError('no inner product defined for '
                                      'exponent != 2 (got {})'
                                      ''.format(self.exponent))

        inners = np.fromiter(
            (x1i.inner(x2i) for x1i, x2i in zip(x1, x2)),
            dtype=x1[0].space.dtype, count=len(x1))

        inner = self.const * np.sum(inners)
        if is_real_dtype(x1[0].dtype):
            return float(inner)
        else:
            return complex(inner)

    def norm(self, x):
        """Calculate the constant-weighted norm of an element.

        Parameters
        ----------
        x1 : `ProductSpaceElement`
            Element whose norm is calculated.

        Returns
        -------
        norm : float
            The norm of the element.
        """
        if self.exponent == 2.0:
            norm_squared = self.inner(x, x).real  # TODO: optimize?!
            return np.sqrt(norm_squared)
        else:
            norms = np.fromiter(
                (xi.norm() for xi in x), dtype=np.float64, count=len(x))

            if self.exponent in (1.0, float('inf')):
                return (self.const *
                        float(np.linalg.norm(norms, ord=self.exponent)))
            else:
                return (self.const ** (1 / self.exponent) *
                        float(np.linalg.norm(norms, ord=self.exponent)))

    def dist(self, x1, x2):
        """Calculate the constant-weighted distance between two elements.

        Parameters
        ----------
        x1, x2 : `ProductSpaceElement`
            Elements whose mutual distance is calculated.

        Returns
        -------
        dist : float
            The distance between the elements.
        """
        if self.dist_using_inner:
            norms1 = np.fromiter(
                (x1i.norm() for x1i in x1),
                dtype=np.float64, count=len(x1))
            norm1 = np.linalg.norm(norms1)

            norms2 = np.fromiter(
                (x2i.norm() for x2i in x2),
                dtype=np.float64, count=len(x2))
            norm2 = np.linalg.norm(norms2)

            inners = np.fromiter(
                (x1i.inner(x2i) for x1i, x2i in zip(x1, x2)),
                dtype=x1[0].space.dtype, count=len(x1))
            inner_re = np.sum(inners.real)

            dist_squared = norm1 ** 2 + norm2 ** 2 - 2 * inner_re
            if dist_squared < 0.0:  # Compensate for numerical error
                dist_squared = 0.0
            return np.sqrt(self.const) * float(np.sqrt(dist_squared))
        else:
            dnorms = np.fromiter(
                ((x1i - x2i).norm() for x1i, x2i in zip(x1, x2)),
                dtype=np.float64, count=len(x1))

            if self.exponent == float('inf'):
                return self.const * np.linalg.norm(dnorms, ord=self.exponent)
            else:
                return (self.const ** (1 / self.exponent) *
                        np.linalg.norm(dnorms, ord=self.exponent))


class ProductSpaceNoWeighting(NoWeightingBase, ProductSpaceConstWeighting):

    """Weighting of `ProductSpace` with constant 1."""

    # Implement singleton pattern for efficiency in the default case
    _instance = None

    def __new__(cls, *args, **kwargs):
        """Implement singleton pattern if ``exp==2.0``."""
        if len(args) == 0:
            exponent = kwargs.pop('exponent', 2.0)
            dist_using_inner = kwargs.pop('dist_using_inner', False)
        elif len(args) == 1:
            exponent = args[0]
            args = args[1:]
            dist_using_inner = kwargs.pop('dist_using_inner', False)
        else:
            exponent = args[0]
            dist_using_inner = args[1]
            args = args[2:]

        if exponent == 2.0 and not dist_using_inner:
            if not cls._instance:
                cls._instance = super().__new__(cls, *args, **kwargs)
            return cls._instance
        else:
            return super().__new__(cls, *args, **kwargs)

    def __init__(self, exponent=2.0, dist_using_inner=False):
        """Initialize a new instance.

        Parameters
        ----------
        exponent : positive float
            Exponent of the norm. For values other than 2.0, the inner
            product is not defined.
        dist_using_inner : bool, optional
            Calculate ``dist`` using the formula

                ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``

            This avoids the creation of new arrays and is thus faster
            for large arrays. On the downside, it will not evaluate to
            exactly zero for equal (but not identical) ``x`` and ``y``.

            Can only be used if ``exponent`` is 2.0.
        """
        super().__init__(impl='numpy', exponent=exponent,
                         dist_using_inner=dist_using_inner)


class ProductSpaceCustomInnerProduct(CustomInnerProductBase):

    """Class for handling a user-specified inner product on `ProductSpace`."""

    def __init__(self, inner, dist_using_inner=False):
        """Initialize a new instance.

        Parameters
        ----------
        inner : callable
            The inner product implementation. It must accept two
            `ProductSpaceElement` arguments, return a element from
            the field of the space (real or complex number) and
            satisfy the following conditions for all space elements
            ``x, y, z`` and scalars ``s``:

            - ``<x, y> = conj(<y, x>)``
            - ``<s*x + y, z> = s * <x, z> + <y, z>``
            - ``<x, x> = 0``  if and only if  ``x = 0``

        dist_using_inner : bool, optional
            Calculate ``dist`` using the formula

                ``||x - y||^2 = ||x||^2 + ||y||^2 - 2 * Re <x, y>``

            This avoids the creation of new arrays and is thus faster
            for large arrays. On the downside, it will not evaluate to
            exactly zero for equal (but not identical) ``x`` and ``y``.

            Can only be used if ``exponent`` is 2.0.
        """
        super().__init__(impl='numpy', inner=inner,
                         dist_using_inner=dist_using_inner)


class ProductSpaceCustomNorm(CustomNormBase):

    """Class for handling a user-specified norm on `ProductSpace`.

    Note that this removes ``inner``.
    """

    def __init__(self, norm):
        """Initialize a new instance.

        Parameters
        ----------
        norm : callable
            The norm implementation. It must accept a
            `ProductSpaceElement` argument, return a float and satisfy
            the following conditions for all space elements
            ``x, y`` and scalars ``s``:

            - ``||x|| >= 0``
            - ``||x|| = 0``  if and only if  ``x = 0``
            - ``||s * x|| = |s| * ||x||``
            - ``||x + y|| <= ||x|| + ||y||``
        """
        super().__init__(norm, impl='numpy')


class ProductSpaceCustomDist(CustomDistBase):

    """Class for handling a user-specified distance on `ProductSpace`.

    Note that this removes ``inner`` and ``norm``.
    """

    def __init__(self, dist):
        """Initialize a new instance.

        Parameters
        ----------
        dist : callable
            The distance function defining a metric on
            `ProductSpace`. It must accept two `ProductSpaceElement`
            arguments and fulfill the following mathematical conditions
            for any three space elements ``x, y, z``:

            - ``dist(x, y) >= 0``
            - ``dist(x, y) = 0``  if and only if  ``x = y``
            - ``dist(x, y) = dist(y, x)``
            - ``dist(x, y) <= dist(x, z) + dist(z, y)``
        """
        super().__init__(dist, impl='numpy')


def _strip_space(x):
    """Strip the SPACE.element( ... ) part from a repr."""
    r = repr(x)
    space_repr = '{!r}.element('.format(x.space)
    if r.startswith(space_repr) and r.endswith(')'):
        r = r[len(space_repr):-1]
    return r


def _indent(x):
    """Indent a string by 4 characters."""
    lines = x.splitlines()
    for i, line in enumerate(lines):
        lines[i] = '    ' + line
    return '\n'.join(lines)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()
